import 'dart:async';
import 'dart:collection';
import 'package:flutter/services.dart';
import 'package:ipda_scanner/scanner_callback.dart';

import 'package:synchronized/synchronized.dart';

export 'package:ipda_scanner/scanner_callback.dart';
export 'package:ipda_scanner/code_format.dart';

class IPdaScanner {
  static const _METHOD_CHANNEL = "ipdascanner";
  static const _SET_PROPERTIES = "setProperties";
  static const _START_SCANNER = "startScanner";
  static const _RESUME_SCANNER = "resumeScanner";
  static const _PAUSE_SCANNER = "pauseScanner";
  static const _STOP_SCANNER = "stopScanner";
  static const _ON_DECODED = "onDecoded";
  static const _ON_ERROR = "onError";

  MethodChannel _channel;
  ScannerCallBack _scannerCallBack;

  final _barcodeQueue = new ListQueue<String>();
  bool _stopProcessing = false;
  bool _isProcessQueueRunning = false;
  bool _useQueue;
  final _scanLock = new Lock(reentrant: true);

  IPdaScanner({ScannerCallBack scannerCallBack, bool useQueue = true}) {
    _channel = const MethodChannel(_METHOD_CHANNEL);
    _channel.setMethodCallHandler(_onMethodCall);
    _scannerCallBack = scannerCallBack;
    _useQueue = useQueue;
    starProcessingLoop();
  }

  set scannerCallBack(ScannerCallBack scannerCallBack) =>
      _scannerCallBack = scannerCallBack;

  setScannerCallBack(ScannerCallBack scannerCallBack) =>
      _scannerCallBack = scannerCallBack;

  Future _onMethodCall(MethodCall call) {
    try {
      switch (call.method) {
        case _ON_DECODED:
          if (_useQueue == false) {
            onDecoded(call.arguments);
          } else {
            _scanLock.synchronized(() => _barcodeQueue.add(call.arguments));
          }

          break;
        case _ON_ERROR:
          onError(Exception(call.arguments));
          break;
        default:
          print(call.arguments);
      }
    } catch (e) {
      print(e);
    }
  }

  ///Called when decoder has successfully decoded the code
  ///<br>
  ///Note that this method always called on a worker thread
  ///
  ///@param code Encapsulates the result of decoding a barcode within an image
  void onDecoded(String code) {
    if (_scannerCallBack != null) _scannerCallBack.onDecoded(code);
  }

  ///Called when error has occurred
  ///<br>
  ///Note that this method always called on a worker thread
  ///
  ///@param error Exception that has been thrown
  void onError(Exception error) {
    if (_scannerCallBack != null) _scannerCallBack.onError(error);
  }

  Future setProperties(Map<String, dynamic> mapProperties) {
    return _channel.invokeMethod(_SET_PROPERTIES, mapProperties);
  }

  Future startScanner() {
    if (_isProcessQueueRunning == false) {
      starProcessingLoop();
    }

    return _channel.invokeMethod(_START_SCANNER);
  }

  Future resumeScanner() {
    return _channel.invokeMethod(_RESUME_SCANNER);
  }

  Future pauseScanner() {
    return _channel.invokeMethod(_PAUSE_SCANNER);
  }

  Future stopScanner() {
    _stopProcessing = true;
    return _channel.invokeMethod(_STOP_SCANNER);
  }

  Future<void> starProcessingLoop() async {
    if (this._useQueue == false) return;

    _stopProcessing = false;
    _isProcessQueueRunning = true;

    while (_stopProcessing == false) {
      var barcode = await _scanLock.synchronized(
          () => _barcodeQueue.isNotEmpty ? _barcodeQueue.removeFirst() : null);

      if (barcode != null) {
        onDecoded(barcode);
      }

      await Future.delayed(Duration(milliseconds: 200));
    }

    _isProcessQueueRunning = false;
  }
}
